/*  DIY Oregon like rain sensor - ATtiny85 - V1.0 by sperate
 *  
 *  RF libarary (slightly modified): 
 *  OS_V3_rainsensor by deberman - https://github.com/deberman/OS_V3_rainsensor
 *  
 *  Wiring (ATtiny85):
 *                            +-------+
 *                           1|*      |8  VCC
 *                           2|       |7  PB2/INT0/PCINT2 (ILS tilt bucket switch)
 *  Data TX RF 433Mhz - PB4  3|       |6  
 *                      GND  4|       |5 
 *                            +-------+
 *
 * 
*/

#include "os_v3_rainsensor.h"
#include "Configuration.h"

OS_v3_rainsensor rainsensor(TX_PIN, ID);    // creates class and defines transmitter pin and ID number
                                            // see Configuration.h
volatile uint16_t bucket = 0;
volatile uint8_t  tilt = false;

inline void mcu_set_power_down_mode()
{
  MCUCR |= (1 << SM1) | (1 << SE);          // select Power-down Mode (DS 7.5.1)
}

inline void mcu_set_watchdog()
{
  MCUSR   &= ~(1 << WDRF);                  // clear the reset flag (DS 8.5.1)
  WDTCR   |=  (1 << WDCE) | (1 << WDE);     // Watchdog Timer Prescale (DS 8.4.1.2 & 8.5.2)
  WDTCR    =  (1 << WDP3) | (1 << WDP0);    // set to roughly 8s (max) (DS 8.5.2, Table 8-3)
  WDTCR   |=  (1 << WDIE);                  // enable interrupt mode
}

inline void mcu_sleep(uint8_t wdt_period)
{
  // Reduce Power Consumption (Low Energy):
  // - ADC off
  // DS 7.4.1 & 17.13.2
  ADCSRA &= ~(1 << ADEN);

  for (uint8_t i = wdt_period; i > 0; i--)
  {
    if (tilt) break;                        // get out of sleep if there was a bucket tilt
    __asm__ __volatile__("sleep");          // in order to send the value immediately
  }
  
  // Re-enable ADC (Analog Digital Converter)
  // needed for measurements with bandgap voltage reference :
  // 25 ADC clock cycles vs 13 if ADC is not re-enabled
  ADCSRA |= (1 << ADEN);
}

inline void mcu_set_interrupts()
{
  GIMSK |=  (1 << PCIE);        // Enables External Interrupt on any PORTB pin change
  PCMSK |=  (1 << PCINT2);      // Enables Interrupt on PB2 (pin7)
  SREG  |=  (1 << SREG_I);      // Set Global Interrupt Enable bit to 1
}

inline int mcu_read_vcc()
{  
  // VCC read value doesn't change unless 
  // we call an analogRead() before reading it
  analogRead(A0);
  
  // Selection de Vbg (select bandgap reference voltage) sur le multiplexeur d'entrée de l'ADC  
  // DS Table 17-4 : cf (2) for choice '1100' for AtTiny  
  ADMUX = (1 << MUX3) | (1 << MUX2);
  
  // After switching to internal voltage reference the ADC requires 
  // a settling time of 1ms before measurements are stable (take 2ms)
  delay(2);
  
  ADCSRA |= (1 << ADSC);         // start conversion (DS 17.3.2)
  while (ADCSRA & (1 << ADSC));  // wait while conversion is in progress
  
  return ((uint32_t) 1024 * 1100) / ((ADCH << 8) | ADCL);
}

void setup(void)
{
  DDRB  &= ~(1 << PORTB2);        // Set ATtiny pin7 (PB2 / INT0 / PCINT2) as input
  PORTB |=  (1 << PORTB2);        // Enable internal pull-up on this pin

  ADCSRA |= ((1 << ADEN) | 6);    // Enable ADC converter | prescaler set to 6 eq. 1/64 (8000kHz/64 = 125kHz)
                                  // ADC freq. must be between 50kHz and 200kHz in order to get max accuracy (DS 17.5)

  mcu_set_watchdog();
  mcu_set_power_down_mode();
  mcu_set_interrupts();
}

void loop(void)
{
  if(tilt){
    delay(1000);                  // Wait for the bucket to finish tilt (2 x "pin chnage" interrupt per tilt)
    tilt = false;
    bucket += 10;
  }
 
  rainsensor.setRainRate(0);
  
  if (bucket > 24000) bucket -= 24000;
  rainsensor.setTotalRain(bucket);          // Pour une valeur envoyée de 10, RFLINK reçoit la valeur 1mm
  rainsensor.setBatteryStatus(mcu_read_vcc() > LOW_BATTERY_ALERT);
  
  rainsensor.buildAndSendPacket();
  
  mcu_sleep(SLEEPING_TIME_S/8);             // par multiple de 8 secondes

}

ISR(WDT_vect)
{}

ISR(PCINT0_vect)
{
  tilt = true;                              // mémorisation de la bascule par interruption externe
}
